/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.ss.formula.functions;

import java.util.Vector;

import org.apache.poi.ss.formula.eval.AreaEvalBase;
import org.apache.poi.ss.formula.eval.BlankEval;
import org.apache.poi.ss.formula.eval.ErrorEval;
import org.apache.poi.ss.formula.eval.EvaluationException;
import org.apache.poi.ss.formula.eval.MissingArgEval;
import org.apache.poi.ss.formula.eval.NotImplementedException;
import org.apache.poi.ss.formula.eval.NumberEval;
import org.apache.poi.ss.formula.eval.OperandResolver;
import org.apache.poi.ss.formula.eval.ValueEval;

/**
 * Implementation for the Excel function SUBTOTAL<p>
 *
 * <b>Syntax :</b> <br/>
 *  SUBTOTAL ( <b>functionCode</b>, <b>ref1</b>, ref2 ... ) <br/>
 *    <table border="1" cellpadding="1" cellspacing="0" summary="Parameter descriptions">
 *      <tr><td><b>functionCode</b></td><td>(1-11) Selects the underlying aggregate function to be used (see table below)</td></tr>
 *      <tr><td><b>ref1</b>, ref2 ...</td><td>Arguments to be passed to the underlying aggregate function</td></tr>
 *    </table><br/>
 * </p>
 *
 *  <table border="1" cellpadding="1" cellspacing="0" summary="Parameter descriptions">
 *      <tr><th>functionCode</th><th>Aggregate Function</th></tr>
 *      <tr align='center'><td>1</td><td>AVERAGE</td></tr>
 *      <tr align='center'><td>2</td><td>COUNT</td></tr>
 *      <tr align='center'><td>3</td><td>COUNTA</td></tr>
 *      <tr align='center'><td>4</td><td>MAX</td></tr>
 *      <tr align='center'><td>5</td><td>MIN</td></tr>
 *      <tr align='center'><td>6</td><td>PRODUCT</td></tr>
 *      <tr align='center'><td>7</td><td>STDEV</td></tr>
 *      <tr align='center'><td>8</td><td>STDEVP *</td></tr>
 *      <tr align='center'><td>9</td><td>AVERAGE</td></tr>
 *      <tr align='center'><td>10</td><td>VAR *</td></tr>
 *      <tr align='center'><td>11</td><td>VARP *</td></tr>
 *      <tr align='center'><td>101-111</td><td>*</td></tr>
 *  </table><br/>
 * * Not implemented in POI yet. Functions 101-111 are the same as functions 1-11 but with
 * the option 'ignore hidden values'.
 * <p/>
 *
 * @author Paul Tomlin &lt; pault at bulk sms dot com &gt;
 */
public class Subtotal implements Function {
	
	public static final int FUNC_SUBTOTAL = 344; //see FunctionEval.produceFunctions();
	public static final int COUNT = 2; 
	public static final int COUNTA = 3; 

	private static Function findFunction(int functionCode) throws EvaluationException {
		switch (functionCode) {
			case 1: return AggregateFunction.AVERAGE;
			case 2: return new Count();
			case 3: return new Counta();
			case 4: return AggregateFunction.MAX;
			case 5: return AggregateFunction.MIN;
			case 6: return AggregateFunction.PRODUCT;
			case 7: return AggregateFunction.STDEV;
			case 8: throw new NotImplementedException("STDEVP");
			case 9: return AggregateFunction.SUM;
			case 10: throw new NotImplementedException("VAR");
			case 11: throw new NotImplementedException("VARP");
		}
		if (functionCode > 100 && functionCode < 112) {
			throw new NotImplementedException("SUBTOTAL - with 'exclude hidden values' option");
		}
		throw EvaluationException.invalidValue();
	}

	public ValueEval evaluate(ValueEval[] args, int srcRowIndex, int srcColumnIndex) {
		int nInnerArgs = args.length-1; // -1: first arg is used to select from a basic aggregate function
		if (nInnerArgs < 1) {
			return ErrorEval.VALUE_INVALID;
		}

		Function innerFunc; 
		int functionCode = -1;
		try {
			ValueEval ve = OperandResolver.getSingleValue(args[0], srcRowIndex, srcColumnIndex);
			functionCode = OperandResolver.coerceValueToInt(ve);
			innerFunc = findFunction(functionCode);
		} catch (EvaluationException e) {
			return e.getErrorEval();
		}

		/*
		 * @author Ulrich Wenzel / IconParc GmbH
		 */
		/*********************************************************/
		//Collect all values of all areas without nested subtotals
		/*********************************************************/
		DoubleList doubleList = new DoubleList();
		int count = 0;
		/************************************************/
		//Iterate over all areas
		/************************************************/
		for( int i=1; i < args.length; i++ ){
			AreaEvalBase lazyArea = (AreaEvalBase)args[i];
			int rowFirst = lazyArea.getFirstRow();
			int rowLast = lazyArea.getLastRow();
			int colFirst = lazyArea.getFirstColumn();
			int colLast = lazyArea.getLastColumn();

			for( int iRow = rowFirst; iRow <= rowLast; iRow++ ){
				for( int iCol = colFirst; iCol <= colLast; iCol++ ){

					int formulaIndex = lazyArea.getFunctionIndex( iRow, iCol );
					if( formulaIndex == FUNC_SUBTOTAL ){
						if( iRow == srcRowIndex  &&  iCol == srcColumnIndex ){ //create Self-Reference-Error
							ValueEval ret = lazyArea.getValue( iRow-rowFirst, iCol-colFirst );
							return ret;
						}
						continue;
					}

					ValueEval ret = lazyArea.getValue( iRow-rowFirst, iCol-colFirst );

					switch( functionCode ){
					case COUNT:
						if( (ret instanceof NumberEval ) ||  ret == MissingArgEval.instance)
							count++; // only missing arguments and numbers are counted
						break;
					case COUNTA:
						if( ret != BlankEval.instance)
							count++;  // Note - everything but BlankEval counts
						break;
					default:
						if( ret instanceof NumberEval )
							doubleList.add( ((NumberEval)ret).getNumberValue() );
						else if( ret instanceof ErrorEval )
							return ret;
					}
				}
			}
		}
		switch( functionCode ){
		case COUNT:
		case COUNTA:
			return new NumberEval( count ); 
		default:
		}
		try {
			double[] values = doubleList.toArray();

			return new NumberEval( ((AggregateFunction) innerFunc).evaluate( values ));

		} catch (EvaluationException e) {
			throw new RuntimeException( e );
		}
	}
	/**
	 * Ulrich Wenzel / IconParc GmbH 2011-06: Taken from MultiOperandNumericFunction
	 */
	private static class DoubleList {
		private double[] _array;
		private int _count;

		public DoubleList() {
			_array = new double[8];
			_count = 0;
		}

		public double[] toArray() {
			if(_count < 1) {
				return MultiOperandNumericFunction.EMPTY_DOUBLE_ARRAY;
			}
			double[] result = new double[_count];
			System.arraycopy(_array, 0, result, 0, _count);
			return result;
		}

		private void ensureCapacity(int reqSize) {
			if(reqSize > _array.length) {
				int newSize = reqSize * 3 / 2; // grow with 50% extra
				double[] newArr = new double[newSize];
				System.arraycopy(_array, 0, newArr, 0, _count);
				_array = newArr;
			}
		}

		public void add(double value) {
			ensureCapacity(_count + 1);
			_array[_count] = value;
			_count++;
		}
	}

}
