/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.ss.formula.functions;

import junit.framework.TestCase;

import org.apache.poi.hssf.HSSFTestDataSamples;
import org.apache.poi.hssf.usermodel.HSSFCell;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellValue;
import org.apache.poi.ss.usermodel.FormulaEvaluator;

/**
 * Tests for {@link Subtotal}
 *
 * @author Paul Tomlin
 */
public final class TestSubtotalNested extends TestCase {
	
	/** Name of the test spreadsheet (found in the standard test data folder) */
	public final static String FILENAME = "SubtotalsNested.xls";


	private static void confirmExpectedResult( FormulaEvaluator evaluator, String msg, Cell cell, double expected ) {
		
		CellValue value = evaluator.evaluate( cell );
		if( value.getErrorValue() != 0 )
			throw new RuntimeException( msg + ": " + value.formatAsString() );
		assertEquals( msg, expected, value.getNumberValue() );
	}
	public void testFunctionsFromTestSpreadsheet() {
		HSSFWorkbook workbook = HSSFTestDataSamples.openSampleWorkbook(FILENAME);
		HSSFSheet sheet = workbook.getSheetAt(0);
		FormulaEvaluator evaluator = workbook.getCreationHelper().createFormulaEvaluator();
		
		assertEquals( "A1", 10.0, sheet.getRow( 0 ).getCell( 0 ).getNumericCellValue());
		assertEquals( "A2", 20.0, sheet.getRow( 1 ).getCell( 0 ).getNumericCellValue() );
		
		//Test simple subtotal over one area
		Cell cellA3 = sheet.getRow( 2 ).getCell( 0 );
		confirmExpectedResult(evaluator, "A3", cellA3, 30.0 );
		
		//Test existence of the second area
		assertNotNull( "B1 must not be null", sheet.getRow( 0 ).getCell( 1 ) );
		assertEquals( "B1", 7.0, sheet.getRow( 0 ).getCell( 1 ).getNumericCellValue() );
		
		Cell cellC1 = sheet.getRow( 0 ).getCell( 2 );
		Cell cellC2 = sheet.getRow( 1 ).getCell( 2 );
		Cell cellC3 = sheet.getRow( 2 ).getCell( 2 );
		
		//Test Functions SUM, COUNT and COUNTA calculation of SUBTOTAL
		//a) areas A and B are used
		//b) first 2 subtotals don't consider the value of nested subtotal in A3
		confirmExpectedResult( evaluator, "SUBTOTAL(SUM;A1:A7;B1:B7)", cellC1, 37.0  );
		confirmExpectedResult( evaluator, "SUBTOTAL(COUNT;A1:A7;B1:B7)", cellC2, 3.0  );
		confirmExpectedResult( evaluator, "SUBTOTAL(COUNTA;A1:A7;B1:B7)", cellC3, 5.0  );
	}
}
